<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\Controller\Adminhtml\Misc;

class Options extends \Magento\Backend\App\Action implements \Magento\Framework\App\Action\HttpGetActionInterface
{
    /**
     * Authorization level of a basic admin session
     */
    public const ADMIN_RESOURCE = 'Mediastrategi_PostNord::options';

    /**
     * @var \Mediastrategi\PostNord\Helper\Meta
     */
    private $metaHelper;

    /**
     * @var \Mediastrategi\PostNord\Helper\Shipment
     */
    private $shipmentHelper;

    /**
     * \Mediastrategi\PostNord\Helper\OrderData
     */
    private $orderDataHelper;

    /**
     * @var \Magento\Sales\Api\OrderRepositoryInterface
     */
    private $orderRepository;

    /**
     * @var \Magento\Framework\Controller\Result\JsonFactory
     */
    private $resultJsonFactory;

    /**
     * @var \Magento\Store\Model\StoreManagerInterface
     */
    private $storeManager;

    public function __construct(
        \Magento\Backend\App\Action\Context $context,
        \Mediastrategi\PostNord\Helper\Meta $metaHelper,
        \Mediastrategi\PostNord\Helper\Shipment $shipmentHelper,
        \Mediastrategi\PostNord\Helper\OrderData $orderDataHelper,
        \Magento\Sales\Api\OrderRepositoryInterface $orderRepository,
        \Magento\Framework\Controller\Result\JsonFactory $resultJsonFactory,
        \Magento\Store\Model\StoreManagerInterface $storeManager
    ) {
        parent::__construct($context);
        $this->metaHelper = $metaHelper;
        $this->shipmentHelper = $shipmentHelper;
        $this->orderDataHelper = $orderDataHelper;
        $this->orderRepository = $orderRepository;
        $this->resultJsonFactory = $resultJsonFactory;
        $this->storeManager = $storeManager;
    }

    /**
     * Execute action based on request and return result
     *
     * @return \Magento\Framework\Controller\ResultInterface
     */
    public function execute(): \Magento\Framework\Controller\ResultInterface
    {
        $result =
            $this->resultJsonFactory->create();
        $orderId =
            (int) $this->getRequest()->getParam('orderId');
        $action =
            (string) $this->getRequest()->getParam('action');
        $data = [];

        switch ($action) {
            case 'refresh':
                $data = $this->refreshAction($orderId);
                break;
            case 'save':
                $data = $this->saveAction($orderId);
                break;
            default:
                $data = ['error' => __('Invalid action!')];
        }

        return $result->setData($data);
    }

    private function getQueryForOrder(
        int $orderId
    ): array {
        $query = [];
        try {
            $order =
                $this->orderRepository->get($orderId);
        } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
            $order = false;
        }
        if ($order) {
            $weightInGrams = $this->shipmentHelper->getOrderShippableWeightInGrams(
                $orderId
            );

            /**
             * Check quote address if we have a company defined
             * if so set type to 'business' otherwise to 'personal'
             */
            $customerType = 'personal';
            foreach ($order->getAddressesCollection() as $address) {
                /** @var \Magento\Quote\Model\Quote\Address $address */
                if ($address->getCompany()) {
                    $customerType = 'business';
                    break;
                }
            }

            $query = [
                'currency' =>
                    (string) $this->storeManager->getStore()
                        ->getBaseCurrency()
                        ->getCurrencyCode(),
                'language' =>
                    (string) $this->metaHelper->getLanguageCode(),
                'recipient' => [
                    'countryCode' =>
                        (string) $order->getShippingAddress()->getCountryId(),
                    'postalCode' =>
                        (string) $order->getShippingAddress()->getPostcode(),
                    'type' =>
                        $customerType,
                ],
                'totalPrice' => (int) round(
                    (float) $order->getSubtotal(),
                    2
                ) * 100,
                'weightInGrams' =>
                    (int) $weightInGrams,
            ];

            $this->metaHelper->log(
                sprintf(
                    '%s calculated query from order: %s to %s',
                    __METHOD__,
                    $orderId,
                    var_export($query, true)
                )
            );
        } else {
            $this->metaHelper->log(
                sprintf(
                    '%s failed to load order with id %s',
                    __METHOD__,
                    $orderId
                )
            );
        }
        return $query;
    }

    /**
     * Return list of available shipment options for order.
     */
    private function refreshAction(int $orderId): array
    {
        $data = [];
        if ($query = $this->getQueryForOrder($orderId)) {
            $data['query'] = $query;
            $shippingOptions = $this->metaHelper->getShippingOptionsForQuery(
                $query,
                (string) $orderId
            );
            $this->metaHelper->log(
                (string) __(
                    'shipping-options: %1',
                    var_export($shippingOptions, true)
                )
            );
            if (!empty($shippingOptions)) {
                /**
                 * Determine the first shipment option id
                 * And build hash-list of available shipment option ids
                 */
                $firstShipmentOptionId =
                    0;
                $availableShipmentOptionIds =
                    [];
                foreach ($shippingOptions as $shippingOption) {
                    if (!$firstShipmentOptionId) {
                        $firstShipmentOptionId =
                            $shippingOption['shipmentOptionId'];
                    }
                    $availableShipmentOptionIds
                    [$shippingOption['shipmentOptionId']] =
                        $shippingOption;
                }

                /**
                 * Try to determine previous shipment option id
                 */
                $previousShipmentOptionId =
                    0;
                try {
                    $order =
                        $this->orderRepository->get($orderId);
                } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
                    $order = false;
                }
                if ($order) {
                    $shippingMethod = $order->getShippingMethod();
                    if (
                        !empty($shippingMethod)
                        && is_string($shippingMethod)
                    ) {
                        $explode = explode(
                            '_',
                            $shippingMethod
                        );
                        $carrierId = $explode[0];
                        $methodId = $explode[1];
                        if (
                            $carrierId === 'mspostnord'
                            && $methodId !== 'fallback'
                            && isset($availableShipmentOptionIds[$methodId])
                        ) {
                            $previousShipmentOptionId = $methodId;
                        }
                    }
                }

                /**
                 * Selected previous shipment option id
                 * if it's available, otherwise select first
                 * shipment option
                 */
                $selectedShipmentOptionId =
                    $previousShipmentOptionId ?: $firstShipmentOptionId;

                /**
                 * Determine first service point id of selected shipment
                 * option (if any) and select it.
                 * Build hash-list of available service points ids
                 * of selected shipment option.
                 */
                $firstServicePointId =
                    0;
                $availableServicePoints =
                    [];
                $option =
                    $availableShipmentOptionIds[$selectedShipmentOptionId];
                if (!empty($option['servicePoints'])) {
                    foreach ($option['servicePoints'] as $servicePoint) {
                        if (!$firstServicePointId) {
                            $firstServicePointId =
                                $servicePoint['servicePointId'];
                        }
                        $availableServicePoints
                        [$servicePoint['servicePointId']] = true;
                    }
                }
                $selectedServicePointId =
                    $firstServicePointId;

                $orderData = [];
                if ($order) {
                    $orderData = $this->orderDataHelper->readData(
                        (int) $order->getEntityId()
                    );
                }

                /**
                 * Try to determine previous service point id (if any)
                 * and if it's available still, select it
                 */
                if (
                    !empty($orderData['servicePointId'])
                    && is_string($orderData['servicePointId'])
                ) {
                    $previousServicePointId =
                        $orderData['servicePointId'];
                    if (isset($availableServicePoints[$previousServicePointId])) {
                        $selectedServicePointId =
                            $previousServicePointId;
                    }
                }

                $additionalServices = [];
                if (
                    !empty($orderData['addons'])
                    && is_array($orderData['addons'])
                ) {
                    $addons = $orderData['addons'];
                    if (!empty($option['additionalServices'])) {
                        $validAdditionalServices = [];
                        foreach ($option['additionalServices'] as $additionalService) {
                            if (!empty($additionalService['id'])) {
                                $validAdditionalServices[$additionalService['id']] =
                                    true;
                            }
                        }

                        foreach ($addons as $addonId => $addon) {
                            if (!empty($validAdditionalServices[$addonId])) {
                                $additionalServices[$addonId] =
                                    $addon;
                            }
                        }
                    }
                }

                $data['session'] = [
                    'additionalServices' =>
                        $additionalServices,
                    'selectedServicePointId' =>
                        $selectedServicePointId,
                    'selectedShipmentOptionId' =>
                        $selectedShipmentOptionId,
                ];
                $data['options'] =
                    $shippingOptions;
            }
        }
        return $data;
    }

    private function saveAction(int $orderId): array
    {
        $updated = 0;
        try {
            $order =
                $this->orderRepository->get($orderId);
        } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
            $order = false;
        }
        if ($order) {
            $session =
                $this->getRequest()->getParam('session');
            if ($session) {
                $this->metaHelper->log(
                    (string) __(
                        '%1 received new state %2',
                        __METHOD__,
                        var_export($session, true)
                    )
                );

                // Load order data
                $orderData = $this->orderDataHelper->readData(
                    $orderId
                );

                // Shipment option id
                if (!empty($session['shipmentOptionId'])) {
                    $shipmentOptionId =
                        (string) $session['shipmentOptionId'];
                    $method = sprintf(
                        'mspostnord_%s',
                        $shipmentOptionId
                    );
                    $order->setShippingMethod(
                        $method
                    );
                    $updated++;
                }

                // Set shipping description
                if (
                    !empty($session['carrierTitle'])
                    && !empty($session['methodTitle'])
                ) {
                    $shippingDescription = sprintf(
                        '%s - %s',
                        (string) $session['carrierTitle'],
                        (string) $session['methodTitle']
                    );
                    $order->setShippingDescription(
                        $shippingDescription
                    );
                    $updated++;
                }

                // Service point id
                if (
                    !empty($session['servicePointId'])
                    && is_string($session['servicePointId'])
                ) {
                    $orderData['servicePointId'] =
                        $session['servicePointId'];
                    $updated++;
                } elseif (!empty($orderData['servicePointId'])) {
                    $orderData['servicePointId'] = '';
                    $updated++;
                }

                // Service point name
                if (
                    !empty($session['servicePointName'])
                    && is_string($session['servicePointName'])
                ) {
                    $orderData['servicePointName'] =
                        $session['servicePointName'];
                    $updated++;
                } elseif (!empty($orderData['servicePointName'])) {
                    $orderData['servicePointName'] = '';
                    $updated++;
                }

                // Selected add-ons
                if (
                    !empty($session['additionalServices'])
                    && is_array($session['additionalServices'])
                ) {
                    $orderData['addons'] =
                        $session['additionalServices'];

                    // Convert stringed boolean to real boolean
                    foreach ($orderData['addons'] as &$addon) {
                        if (
                            !empty($addon['value'])
                            && $addon['value'] === 'true'
                        ) {
                            $addon['value'] = true;
                        }
                    }
                    $updated++;
                } elseif (!empty($orderData['addons'])) {
                    $orderData['addons'] = [];
                    $updated++;
                }

                if ($updated) {
                    $this->orderDataHelper->update(
                        $orderId,
                        $orderData
                    );
                    $order->save();

                    if (
                        !empty($orderData['shipments'])
                        && is_array($orderData['shipments'])
                    ) {
                        foreach ($orderData['shipments'] as $shipment) {
                            if (
                                !empty($shipment['id'])
                                && is_string($shipment['id'])
                                && !empty($shipment['status'])
                                && is_string($shipment['status'])
                                && $shipment['status']
                                === \Mediastrategi\PostNord\Helper\Api::SHIPMENT_STATUS_RESERVED
                            ) {
                                $this->shipmentHelper->updateOrderShipmentReservation(
                                    $orderId,
                                    $shipment['id'],
                                    $session['shipmentOptionId'],
                                    (!empty($session['servicePointId'])
                                        && is_string($session['servicePointId'])
                                        ? $session['servicePointId']
                                        : ''),
                                    (!empty($session['additionalServices'])
                                        && is_array($session['additionalServices'])
                                        ? $session['additionalServices']
                                        : [])
                                );
                            }
                        }
                    }
                }

                $this->metaHelper->log(
                    (string) __(
                        '%1 updated %1 fields on order',
                        __METHOD__,
                        $updated
                    )
                );
            } else {
                $this->metaHelper->log(
                    (string) __(
                        '%1 received no new state',
                        __METHOD__
                    )
                );
            }
        }
        return ['updated' => $updated];
    }
}
