<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\Helper;

class Api extends \Magento\Framework\App\Helper\AbstractHelper
{
    /**
     * @var \Magento\Framework\Filesystem\DriverInterface
     */
    private $driverInterface;

    private const BASE_URI_PRODUCTION =
        'https://api2.postnord.com/rest/shipment/v1/ostecom';

    private const BASE_URI_TEST =
        'https://atportal.postnord.com/api/ecom/integrator';

    private const IS_PRODUCTION_API_ENVIRONMENT =
        true;

    /**
     * @var array<int, \Mediastrategi\PostNordAPI\Connection>
     */
    private static $storeConnections = [];

    /**
     * @var \Magento\Store\Model\StoreManagerInterface
     */
    private $storeManager;

    /**
     * @var \Mediastrategi\PostNord\Helper\Cache
     */
    private $cacheHelper;

    public const SHIPMENT_STATUS_BOOKED = 'booked';
    public const SHIPMENT_STATUS_FAILED = 'failed';
    public const SHIPMENT_STATUS_ONGOING = 'ongoing';
    public const SHIPMENT_STATUS_RESERVED = 'reserved';

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function updateReservedShipment(
        string $shipmentId,
        array $shipmentData
    ): \Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->putReservedShipments(
                $shipmentId,
                $shipmentData
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Library instantiated the wrong way, contact developers!'
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\General $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected error when performing request: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\Curl $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected cURL error when performing request: %1',
                    $e->getMessage()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function deleteReservedShipment(
        string $shipmentId
    ): \Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->deleteReservedShipments(
                $shipmentId
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Library instantiated the wrong way, contact developers!'
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\General $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected error when performing request: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\Curl $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected cURL error when performing request: %1',
                    $e->getMessage()
                )
            );
        }
    }

    public function __construct(
        \Magento\Framework\App\Helper\Context $context,
        \Magento\Framework\Filesystem\DriverInterface $driverInterface,
        \Magento\Store\Model\StoreManagerInterface $storeManager,
        \Mediastrategi\PostNord\Helper\Cache $cacheHelper
    ) {
        parent::__construct($context);
        $this->driverInterface = $driverInterface;
        $this->storeManager = $storeManager;
        $this->cacheHelper = $cacheHelper;
        $this->autoloadLibraries();
    }

    public function __destruct()
    {
        if (self::$storeConnections) {
            foreach (self::$storeConnections as $storeId => $connection) {
                $apiIdentifier = $this->getConfig(
                    'carriers/mspostnord/api/identifier',
                    $storeId
                );
                $apiKey = $this->getConfig(
                    'carriers/mspostnord/api/key',
                    $storeId
                );
                $apiToken =
                    $connection->getToken();
                $this->cacheHelper->saveApiToken(
                    $apiIdentifier,
                    $apiKey,
                    $apiToken
                );
            }
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getShippingOptionsForQuery(
        array $query
    ): ?\Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->postShippingOptionsQuery(
                $query
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected implementation error, contact developer of module! Errors: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidResponse $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response from server: %1',
                    $e->getMessage()
                )
            );
        } catch (
            \Mediastrategi\PostNordAPI\Exceptions\General
            | \Mediastrategi\PostNordAPI\Exceptions\Curl $e
        ) {
            // Detect wrong credentials here
            if (
                stripos(
                    $e->getMessage(),
                    'invalid credentials'
                ) !== false
            ) {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Invalid credentials or expired auth token!'
                    )
                );
            }

            // Gracefully handle case were API is down
            return null;
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function bookShipments(
        array $shipmentList
    ): \Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->postReservedShipmentsBook(
                $shipmentList
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Library instantiated the wrong way, contact developers!'
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\General $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected error when performing request: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\Curl $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected cURL error when performing request: %1',
                    $e->getMessage()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function printShipment(
        string $shipmentId
    ): \Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->postShipments(
                $shipmentId
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Library instantiated the wrong way, contact developers!'
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\General $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected error when performing request: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\Curl $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected cURL error when performing request: %1',
                    $e->getMessage()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getShipments(
        array $shipmentList
    ): \Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->getShipmentsByShipmentIds(
                implode(
                    ',',
                    $shipmentList
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Library instantiated the wrong way, contact developers!'
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\General $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected error when performing request: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\Curl $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected cURL error when performing request: %1',
                    $e->getMessage()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function reserveShipment(
        array $shipmentData
    ): \Mediastrategi\PostNordAPI\Transaction {
        try {
            return $this->getConnection()->postReservedShipments(
                $shipmentData
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\InvalidArgument $e) {
            $this->log($e->getMessage());
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Library instantiated the wrong way, contact developers! See log for more information'
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\General $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected error when performing request: %1',
                    $e->getMessage()
                )
            );
        } catch (\Mediastrategi\PostNordAPI\Exceptions\Curl $e) {
            throw new \Mediastrategi\PostNordAPI\Exceptions\General(
                __(
                    'Unexpected cURL error when performing request: %1',
                    $e->getMessage()
                )
            );
        }
    }

    private function autoloadLibraries(): void
    {
        $path = $this->driverInterface->getRealPath(
            $this->driverInterface->getParentDirectory(
                __DIR__
            )
        );
        $autoloader =
            \Magento\Framework\Autoload\AutoloaderRegistry::getAutoloader();
        $autoloader->addPsr4(
            'Mediastrategi\\PostNordAPI\\',
            $path . '/Libraries/library-postnord/src/'
        );
    }

    private function getConnection(): \Mediastrategi\PostNordAPI\Connection
    {
        $storeId =
            $this->storeManager->getStore()->getId();
        if (!isset(self::$storeConnections[$storeId])) {
            $paperSize = $this->getConfig(
                'carriers/mspostnord/print_options/paper_size'
            );
            if (!$paperSize) {
                $paperSize = 'A4';
            }

            $labelSize = $this->getConfig(
                'carriers/mspostnord/print_options/label_size'
            );
            if (!$labelSize) {
                $labelSize = 'standard';
            }

            $config = [
                'baseUri' =>
                    self::IS_PRODUCTION_API_ENVIRONMENT
                    ? self::BASE_URI_PRODUCTION
                    : self::BASE_URI_TEST,
                'identifier' => $this->getConfig(
                    'carriers/mspostnord/api/identifier'
                ),
                'key' => $this->getConfig(
                    'carriers/mspostnord/api/key'
                ),
                'labelSize' => $labelSize,
                'paperSize' => $paperSize,
            ];
            if (
                $apiToken = $this->cacheHelper->loadApiToken(
                    $config['identifier'],
                    $config['key']
                )
            ) {
                $config['token'] =
                    $apiToken;
            }
            self::$storeConnections[$storeId] =
                new \Mediastrategi\PostNordAPI\Connection($config);
        }
        return self::$storeConnections[$storeId];
    }

    /**
     * NOTE We cannot use this method from Meta because
     * that would create a circular dependency
     */
    private function getConfig(
        string $path,
        int $storeId = null
    ) {
        try {
            $value = $this->scopeConfig->getValue(
                $path,
                \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
                $storeId ?? $this->storeManager->getStore()->getId()
            );
            return $value;
        } catch (\Exception $e) {
            return '';
        }
    }

    /**
     * NOTE We cannot use this method from Meta because
     * that would create a circular dependency
     */
    public function log(string $message): void
    {
        $this->_logger->debug($message);
    }
}
