<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\Helper;

class Shipment extends \Magento\Framework\App\Helper\AbstractHelper
{
    const TRACKING_LINK_FORMAT =
        'https://portal.postnord.com/tracking/details/%s';

    /**
     * @var \Magento\Framework\Filesystem\DriverInterface
     */
    private $filesystemInterface;

    /**
     * @var \Mediastrategi\PostNord\Helper\Meta
     */
    private $metaHelper;

    /**
     * \Mediastrategi\PostNord\Helper\OrderData
     */
    private $orderDataHelper;

    /**
     * @var \Magento\Sales\Api\OrderItemRepositoryInterface
     */
    private $orderItemRepository;

    /**
     * @var \Magento\Sales\Api\OrderRepositoryInterface
     */
    private $orderRepository;

    /**
     * @var \Magento\Catalog\Model\ProductRepository $productRepository
     */
    private $productRepository;

    /**
     * @var \Magento\Sales\Api\ShipmentRepositoryInterface
     */
    private $shipmentRepository;

    /**
     * @var int
     */
    public $flag = 0;

    public function __construct(
        \Magento\Framework\App\Helper\Context $context,
        \Magento\Framework\Filesystem\DriverInterface $filesystemInterface,
        \Mediastrategi\PostNord\Helper\OrderData $orderDataHelper,
        \Magento\Sales\Api\OrderRepositoryInterface $orderRepository,
        \Magento\Catalog\Model\ProductRepository $productRepository,
        \Magento\Sales\Api\ShipmentRepositoryInterface $shipmentRepository,
        \Magento\Sales\Api\OrderItemRepositoryInterface $orderItemRepository,
        \Mediastrategi\PostNord\Helper\Meta $metaHelper
    ) {
        parent::__construct($context);
        $this->filesystemInterface = $filesystemInterface;
        $this->orderDataHelper = $orderDataHelper;
        $this->orderItemRepository = $orderItemRepository;
        $this->orderRepository = $orderRepository;
        $this->productRepository = $productRepository;
        $this->metaHelper = $metaHelper;
        $this->shipmentRepository = $shipmentRepository;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function getShippingLabelDummyPdf(): string
    {
        $path = $this->filesystemInterface->getParentDirectory(
            $this->filesystemInterface->getParentDirectory(
                __FILE__
            )
        ) . '/Libraries/shipping_label_dummy.pdf';
        if (file_exists($path)) {
            return file_get_contents($path);
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Failed to get shipping label dummy at %1',
                    $path
                )
            );
        }
    }

    public function getExternalShipmentTrackingLink(string $externalShipmentId): string
    {
        try {
            $status = $this->metaHelper->shipmentStatus(
                $externalShipmentId
            );
            foreach ($status['shipments'] as $shipment) {
                if ($shipment['shipmentId'] === $externalShipmentId) {
                    if (!empty($shipment['trackingIds'])) {
                        return sprintf(
                            self::TRACKING_LINK_FORMAT,
                            current($shipment['trackingIds'])
                        );
                    }
                }
            }
        } catch (\Exception $e) {
            return sprintf(
                self::TRACKING_LINK_FORMAT,
                $externalShipmentId
            );
        }
        return '';
    }

    /**
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    public function getShipmentTrackingLink(string $shipmentId): string
    {
        /* @throws \Magento\Framework\Exception\NoSuchEntityException */
        $shipment =
            $this->shipmentRepository->get($shipmentId);
        $trackingId = '';
        if ($tracks = $shipment->getTracks()) {
            /** @var \Magento\Sales\Api\Data\ShipmentTrackInterface $tracks */
            foreach ($tracks as $track) {
                if (
                    $track->getTrackNumber()
                ) {
                    $trackingId =
                        $track->getTrackNumber();
                    break;
                }
            }
        }
        if (!$trackingId) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Order %1 shipment %2 lacks tracking number!',
                    $shipment->getOrderId(),
                    $shipmentId
                )
            );
        }

        $orderData = $this->orderDataHelper->readData(
            $shipment->getOrderId()
        );
        if (empty($orderData['shipments'][$trackingId]['trackingIds'])) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Order %1 shipment %2  lacks tracking ids!',
                    $shipment->getOrderId(),
                    $trackingId
                )
            );
        }

        return sprintf(
            self::TRACKING_LINK_FORMAT,
            current($orderData['shipments'][$trackingId]['trackingIds'])
        );
    }

    /**
     * Do shipment request to carrier web service,
     * obtain Print Shipping Labels and process errors in response
     *
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function processOrderShipmentRequest(
        \Magento\Framework\DataObject $request
    ): \Magento\Framework\DataObject {
        if (!$this->metaHelper->hasProperUsage()) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'You need valid license for PostNord Ship Connect to process PostNord shipment for orders!'
                )
            );
        }
        $orderId = (int) $request->getOrderShipment()->getOrderId();
        try {
            $order = $this->orderRepository->get($orderId);
        } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Failed to load order for shipment! %1',
                    $orderId
                )
            );
        }

        // Try to get order shipment option id
        $shipmentOptionId = 0;
        $shippingMethod = $order->getShippingMethod();
        if (
            !empty($shippingMethod)
            && is_string($shippingMethod)
        ) {
            $explode = explode(
                '_',
                $shippingMethod
            );
            $carrierId = $explode[0];
            $methodId = $explode[1];
            if (
                $carrierId === 'mspostnord'
                && $methodId !== 'fallback'
            ) {
                $shipmentOptionId = $methodId;
            }
        }
        if (!$shipmentOptionId) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Missing required shipment option id on order! %1',
                    $orderId
                )
            );
        }

        $orderData =
            $this->orderDataHelper->readData($orderId);

        $selectedServicePointId = '';
        if (
            !empty($orderData['servicePointId'])
            && is_string($orderData['servicePointId'])
        ) {
            $selectedServicePointId =
                $orderData['servicePointId'];
        }

        $selectedAddons =
            [];
        if (
            !empty($orderData['addons'])
            && is_array($orderData['addons'])
        ) {
            $selectedAddons =
                $orderData['addons'];
        }

        $shipmentData = $this->generateShipmentDataForOrderShipmentRequest(
            $order,
            $request,
            $shipmentOptionId,
            $selectedServicePointId,
            $selectedAddons
        );

        $processMode = $this->metaHelper->getConfig(
            'carriers/mspostnord/process_mode'
        );

        switch ($processMode) {
            case \Mediastrategi\PostNord\Helper\Meta::ORDER_PROCESSING_BOOK:
                return $this->createBookedOrderShipment(
                    $orderId,
                    $shipmentData
                );
            case \Mediastrategi\PostNord\Helper\Meta::ORDER_PROCESSING_RESERVE:
                return $this->createReservedOrderShipment(
                    $orderId,
                    $shipmentData
                );
            default:
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Invalid process-mode %1!',
                        $processMode
                    )
                );
        }
    }

    private function createBookedOrderShipment(
        int $orderId,
        array $shipmentData
    ): \Magento\Framework\DataObject {
        // Reserve shipment
        $decoded = $this->metaHelper->createOrderReservedShipment(
            $orderId,
            $shipmentData
        );
        $shipmentId =
            $decoded['reservedShipment']['shipmentId'];

        // Book shipment
        $decoded = $this->metaHelper->createOrderBookedShipment(
            $orderId,
            $shipmentId
        );
        // NOTE We always book just one so it's always the first one here
        $bookedShipment =
            current($decoded['shipments']);

        // Update order data
        $orderData =
            $this->orderDataHelper->readData($orderId);
        if (empty($orderData['shipments'])) {
            $orderData['shipments'] = [];
        }
        $orderData['shipments'][$bookedShipment['shipmentId']] = [
            'id' =>
                $bookedShipment['shipmentId'],
            'status' =>
                $bookedShipment['status'],
            'trackingIds' =>
                [],
        ];
        $this->orderDataHelper->update(
            $orderId,
            $orderData
        );
        $trackingIds =
            !empty($bookedShipment['trackingIds'])
            ? $bookedShipment['trackingIds']
            : [];

        $isBooked =
            $bookedShipment['status'] ===
            Api::SHIPMENT_STATUS_BOOKED;

        while (!$isBooked) {
            sleep(1);
            $decoded = $this->shipmentStatus(
                $orderId,
                $shipmentId,
                false
            );
            $shipment =
                current($decoded['shipments']);
            switch ($shipment['status']) {
                case Api::SHIPMENT_STATUS_FAILED:
                    throw new \Magento\Framework\Exception\LocalizedException(
                        __(
                            'Failed to book shipment %1!',
                            $shipmentId
                        )
                    );
                case Api::SHIPMENT_STATUS_RESERVED:
                    throw new \Magento\Framework\Exception\LocalizedException(
                        __(
                            'Failed to start booking of shipment %1!',
                            $shipmentId
                        )
                    );
                case Api::SHIPMENT_STATUS_BOOKED:
                    $isBooked =
                        true;
                    $trackingIds =
                        $shipment['trackingIds'];
                    break;
            }
        }

        // Print shipment label
        $shippingLabel = $this->metaHelper->printShipment(
            $orderId,
            $shipmentId
        );

        // Update order data
        $orderData =
            $this->orderDataHelper->readData($orderId);
        $orderData['shipments'][$shipmentId]['trackingIds'] =
            $trackingIds;
        $orderData['shipments'][$shipmentId]['printedLabel'] =
            true;
        $this->orderDataHelper->update(
            $orderId,
            $orderData
        );

        $result = new \Magento\Framework\DataObject();
        if (!empty($trackingIds[0])) {
            $result->setTrackingNumber($trackingIds[0]);
        } else {
            $result->setTrackingNumber($shipmentId);
        }
        $result->setShippingLabelContent($shippingLabel);
        return $result;
    }

    private function createReservedOrderShipment(
        int $orderId,
        array $shipmentData
    ): \Magento\Framework\DataObject {
        $decoded = $this->metaHelper->createOrderReservedShipment(
            $orderId,
            $shipmentData
        );

        // Set order shipment id and status here
        $orderData =
            $this->orderDataHelper->readData($orderId);
        if (empty($orderData['shipments'])) {
            $orderData['shipments'] = [];
        }
        $orderData['shipments'][$decoded['reservedShipment']['shipmentId']] = [
            'id' =>
                $decoded['reservedShipment']['shipmentId'],
            'status' =>
                $decoded['reservedShipment']['status'],
        ];
        $this->orderDataHelper->update(
            $orderId,
            $orderData
        );

        $result = new \Magento\Framework\DataObject();
        $result->setTrackingNumber(
            $decoded['reservedShipment']['shipmentId']
        );
        $result->setShippingLabelContent(
            $this->getShippingLabelDummyPdf()
        );
        return $result;
    }

    public function shipmentStatus(
        int $orderId,
        string $shipmentId,
        bool $printIfPossible
    ): array {
        return $this->shipmentStatuses(
            [
                [
                    $orderId,
                    $shipmentId
                ]
            ],
            $printIfPossible
        );
    }

    public function shipmentStatuses(
        $ordersAndShipments,
        bool $printIfPossible
    ): array {
        $orderIds = [];
        $shipmentIds = [];
        $shipmentIdToKey = [];
        foreach ($ordersAndShipments as $orderAndShipment) {
            $orderIds[] =
                $orderAndShipment[0];
            $shipmentIdToKey[$orderAndShipment[1]] =
                count($shipmentIds);
            $shipmentIds[] =
                $orderAndShipment[1];
        }
        $decoded = $this->metaHelper->orderShipmentsStatuses(
            $ordersAndShipments
        );

        foreach ($decoded['shipments'] as $shipment) {
            $ref =
                $shipmentIdToKey[$shipment['shipmentId']];
            $orderId =
                $orderIds[$ref];
            unset($orderIds[$ref]);
            $orderData =
                $this->orderDataHelper->readData($orderId);
            $orderData['shipments'][$shipment['shipmentId']]['status'] =
                $shipment['status'];
            $this->orderDataHelper->update(
                $orderId,
                $orderData
            );

            if (
                $printIfPossible
                && $shipment['status'] === Api::SHIPMENT_STATUS_BOOKED
                && empty($orderData['printedLabel'])
            ) {
                // Print shipment label
                try {
                    $shippingLabel = $this->metaHelper->printShipment(
                        $orderId,
                        $shipment['shipmentId']
                    );
                } catch (\Exception $e) {
                    // Shipments that have already been delivered cannot be printed
                    $this->metaHelper->log(
                        (string) __(
                            'Could not print label for order %1 and shipment %2, error: %3',
                            $orderId,
                            $shipment['shipmentId'],
                            $e->getMessage()
                        )
                    );

                    // Update order data
                    $orderData =
                        $this->orderDataHelper->readData($orderId);
                    $orderData['shipments'][$shipment['shipmentId']]['printedLabel'] =
                        true;
                    $orderData['shipments'][$shipment['shipmentId']]['printedLabelFailure'] =
                        true;
                    $this->orderDataHelper->update(
                        $orderId,
                        $orderData
                    );
                    continue;
                }

                // Update label here
                $internalShipment = $this->getOrderShipmentFromExternalShipmentId(
                    $orderId,
                    $shipment['shipmentId']
                );
                if (!$internalShipment) {
                    // Update order meta data here
                    $updated = 0;
                    $orderData =
                        $this->orderDataHelper->readData($orderId);
                    if (!empty($orderData['shipments'][$shipment['shipmentId']])) {
                        unset($orderData['shipments'][$shipment['shipmentId']]);
                        $updated++;
                    }
                    if ($updated) {
                        $this->orderDataHelper->update(
                            $orderId,
                            $orderData
                        );
                    }

                    $this->metaHelper->log(
                        (string) __(
                            'Order %1 internal shipment for %2 could not be found',
                            $orderId,
                            $shipment['shipmentId']
                        )
                    );
                    continue;
                }

                $internalShipment->setShippingLabel($shippingLabel);
                $this->shipmentRepository->save($internalShipment);

                // Update order data
                $orderData =
                    $this->orderDataHelper->readData($orderId);
                $orderData['shipments'][$shipment['shipmentId']]['printedLabel'] =
                    true;
                $orderData['shipments'][$shipment['shipmentId']]['trackingIds'] =
                    $shipment['trackingIds'];
                $this->orderDataHelper->update(
                    $orderId,
                    $orderData
                );
            }
        }

        if ($orderIds) {
            /**
             * In case a shipment can't be found anymore on PostNord
             * delete it's order shipment from Magento 2 as well
             */
            foreach ($orderIds as $i => $orderId) {
                // Find Magento 2 associated order shipment here
                $shipmentId = $shipmentIds[$i];
                $this->deleteShipment(
                    $orderId,
                    $shipmentId
                );
            }
        }
        return $decoded;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    private function deleteShipment(
        int $orderId,
        string $externalShipmentId
    ): void {
        $shipment = $this->getOrderShipmentFromExternalShipmentId(
            $orderId,
            $externalShipmentId
        );
        /** @var \Magento\Sales\Model\ResourceModel\Order\Shipment $shipment */
        /* @throws \Magento\Framework\Exception\NoSuchEntityException */
        $order = $this->orderRepository->get(
            $orderId
        );

        if (!$shipment) {
            $this->metaHelper->log(
                (string) __(
                    'No internal shipment can be found skipping deletion.',
                    $orderId,
                    $externalShipmentId
                )
            );
        } else {
            if ($packages = $shipment->getPackages()) {
                foreach ($packages as $package) {
                    if (
                        !empty($package['items'])
                        && is_array($package['items'])
                    ) {
                        foreach ($package['items'] as $item) {
                            if (!empty($item['order_item_id'])) {
                                /* @throws \Magento\Framework\Exception\NoSuchEntityException */
                                $orderItem = $this->orderItemRepository->get(
                                    $item['order_item_id']
                                );
                                if (!$orderItem->getIsVirtual()) {
                                    $orderItem->setQtyShipped(
                                        $orderItem->getQtyShipped()
                                        - $item['qty']
                                    );
                                    $orderItem->save();
                                }
                            }
                        }
                    }
                }
            }

            // Delete order shipment
            $this->shipmentRepository->delete(
                $shipment
            );
        }

        // Update order meta data here
        $updated = 0;
        $orderData =
            $this->orderDataHelper->readData($orderId);
        if (!empty($orderData['shipments'][$externalShipmentId])) {
            unset($orderData['shipments'][$externalShipmentId]);
            $updated++;
        }
        if ($updated) {
            $this->orderDataHelper->update(
                $orderId,
                $orderData
            );
        }

        // Flag that order can ship again
        $order->setActionFlag(
            \Magento\Sales\Model\Order::ACTION_FLAG_SHIP,
            true
        );
        $this->flag = 1;
    }

    public function getOrderShipmentFromExternalShipmentId(
        int $orderId,
        string $externalShipmentId
    ): ?\Magento\Sales\Model\Order\Shipment {
        try {
            $order = $this->orderRepository->get(
                $orderId
            );
        } catch (\Exception $e) {
            $order = false;
        }
        if ($order) {
            $shipments =
                $order->getShipmentsCollection();
            /** @var \Magento\Sales\Model\ResourceModel\Order\Shipment\CollectionFactory $shipments */
            $foundShipment =
                false;
            foreach ($shipments as $shipment) {
                /** @var \Magento\Sales\Model\ResourceModel\Order\Shipment $shipment */
                if ($tracks = $shipment->getTracks()) {
                    /** @var \Magento\Sales\Api\Data\ShipmentTrackInterface $tracks */
                    foreach ($tracks as $track) {
                        if (
                            $track->getTrackNumber()
                            == $externalShipmentId
                        ) {
                            $foundShipment = $shipment;
                            break 2;
                        }
                    }
                }
            }
            if ($foundShipment) {
                return $foundShipment;
            }
        }
        return null;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function generateShipmentDataForOrderShipmentRequest(
        \Magento\Sales\Model\Order $order,
        \Magento\Framework\DataObject $request,
        string $shipmentOptionId,
        string $selectedServicePointId,
        array $selectedAddons
    ): array {
        // Calculate total weight for package
        $requestPackageId =
            (int) $request->getPackageId();
        $packageWeight =
            0.0;
        $weightUnit =
            '';
        if ($packages = $request->getOrderShipment()->getPackages()) {
            foreach ($packages as $packageId => $package) {
                if (
                    $packageId === $requestPackageId
                ) {
                    if (isset($package['params']['weight'])) {
                        $packageWeight +=
                            (float) $package['params']['weight'];
                    }
                    if (isset($package['params']['weight_units'])) {
                        $weightUnit =
                            (string) $package['params']['weight_units'];
                    }
                }
            }
        }
        if (!$weightUnit) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Missing package weight unit!'
                )
            );
        }

        // Convert package-weight to grams
        $weightInGrams = 0;
        $packageWeight =
            (float) $request->getPackageWeight();
        if ($weightUnit === 'POUND') {
            $weightInGrams =
                $packageWeight * 453.5924;
        } elseif ($weightUnit === 'KILOGRAM') {
            $weightInGrams =
                $packageWeight * 1000;
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected weight unit %1!',
                    $weightUnit
                )
            );
        }

        // Build initial data
        $shipmentData = [
            'currency' =>
                (string) $order->getOrderCurrencyCode(),
            'recipient' => [
                'address1' =>
                    (string) $request->getRecipientAddressStreet1(),
                'address2' =>
                    (string) $request->getRecipientAddressStreet2(),
                'city' =>
                    (string) $request->getRecipientAddressCity(),
                'companyName' =>
                    (string) $request->getRecipientContactCompanyName(),
                'countryCode' =>
                    (string) $request->getRecipientAddressCountryCode(),
                'email' =>
                    (string) $request->getRecipientEmail(),
                'name' =>
                    (string) $request->getRecipientContactPersonName(),
                'phone' =>
                    (string) $request->getRecipientContactPhoneNumber(),
                'postalCode' =>
                    (string) $request->getRecipientAddressPostalCode(),
                'type' =>
                    ($request->getRecipientContactCompanyName()
                        ? 'business' : 'personal'),
            ],
            'reference' => sprintf(
                'M2 #%s',
                $order->getIncrementId()
            ),
            'totalPrice' => (int) round(
                (float) $order->getBaseSubtotalInclTax(),
                2
            ) * 100,
            'weightInGrams' =>
                (int) $weightInGrams,
        ];

        // Remove empty recipient fields
        $newShipmentData = $shipmentData;
        foreach ($shipmentData['recipient'] as $key => $data) {
            if (empty($data)) {
                unset($newShipmentData['recipient'][$key]);
            }
        }
        $shipmentData = $newShipmentData;

        // Build data for shipment option id
        $selectedShippingOption = [
            'shipmentOptionId' =>
                (string) $shipmentOptionId,
        ];

        // Build data for selected add-ons
        if (!empty($selectedAddons)) {
            $shipmentDataSelectedAddons =
                [];
            foreach ($selectedAddons as $selectedAddonKey => $selectedAddonValue) {
                if (
                    !empty($selectedAddonValue['value'])
                    && is_string($selectedAddonValue['value'])
                ) {
                    $shipmentDataSelectedAddons[] = [
                        'data' => [
                            'text' =>
                                $selectedAddonValue['value'],
                        ],
                        'id' =>
                            (string) $selectedAddonKey,
                        'selected' =>
                            true,
                    ];
                } else {
                    $shipmentDataSelectedAddons[] = [
                        'id' =>
                            (string) $selectedAddonKey,
                        'selected' =>
                            true,
                    ];
                }
            }
            $selectedShippingOption['additionalServices'] =
                $shipmentDataSelectedAddons;
        }

        // Build data for selected service point id
        if (!empty($selectedServicePointId)) {
            $selectedShippingOption['servicePointId'] =
                $selectedServicePointId;
        }

        $shipmentData['selectedShippingOption'] =
            $selectedShippingOption;

        return $shipmentData;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getOrderShippableWeightInGrams(
        int $orderId
    ): int {
        try {
            $order = $this->orderRepository->get($orderId);
        } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Could not find order %1',
                    $orderId
                )
            );
        }

        // Calculate weight
        $packageWeight = 0;
        foreach ($order->getAllItems() as $orderItem) {
            if ($orderItem->getIsVirtual()) {
                continue;
            }
            try {
                $product = $this->productRepository->getById(
                    $orderItem->getProductId()
                );
            } catch (\Exception $e) {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Could not find product %1',
                        $orderItem->getProductId()
                    )
                );
            }
            $packageWeight +=
                ($product->getWeight() * $orderItem->getQtyOrdered());
        }

        // Set minimum weight here
        $minimumWeight = (float) $this->metaHelper->getConfig(
            'carriers/mspostnord/weight/products_minimum_weight'
        );
        if ($packageWeight < $minimumWeight) {
            $packageWeight = $minimumWeight;
        }

        // Add wrapping weight here
        $wrappingWeight = (float) $this->metaHelper->getConfig(
            'carriers/mspostnord/weight/wrapping_weight'
        );
        $packageWeight += $wrappingWeight;

        // Convert to grams here
        $weightUnit = $this->metaHelper->getConfig(
            'general/locale/weight_unit'
        );
        if ($weightUnit === 'lbs') {
            $weightInGrams =
                $packageWeight * 453.5924;
        } elseif ($weightUnit === 'kgs') {
            $weightInGrams =
                $packageWeight * 1000;
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected weight unit %1!',
                    $weightUnit
                )
            );
        }
        return (int) $weightInGrams;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function generateShipmentDataForOrder(
        int $orderId,
        string $shipmentOptionId,
        string $selectedServicePointId,
        array $selectedAddons
    ): array {
        try {
            $order =
                $this->orderRepository->get($orderId);
        } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Failed to load order for shipment! %1',
                    $orderId
                )
            );
        }

        $weightInGrams = $this->getOrderShippableWeightInGrams(
            $orderId
        );

        // Build initial data
        $shipmentData = [
            'recipient' => [
                'address1' =>
                    (string) $order->getShippingAddress()->getStreetLine(1),
                'address2' =>
                    (string) $order->getShippingAddress()->getStreetLine(2),
                'city' =>
                    (string) $order->getShippingAddress()->getCity(),
                'countryCode' =>
                    (string) $order->getShippingAddress()->getCountryId(),
                'email' =>
                    (string) $order->getShippingAddress()->getEmail(),
                'name' =>
                    (string) ($order->getShippingAddress()->getCompany()
                        ? $order->getShippingAddress()->getCompany()
                        : $order->getShippingAddress()->getName()),
                'phone' =>
                    (string) $order->getShippingAddress()->getTelephone(),
                'postalCode' =>
                    (string) $order->getShippingAddress()->getPostcode(),
                'type' =>
                    ($order->getShippingAddress()->getCompany()
                        ? 'business' : 'personal'),
            ],
            'reference' => sprintf(
                'M2 #%s',
                $order->getIncrementId()
            ),
            'totalPrice' =>
                (int) round(
                    (float) $order->getSubtotal(),
                    2
                )
                * 100,
            'weightInGrams' =>
                (int) $weightInGrams,
        ];

        // Remove empty recipient fields
        $newShipmentData = $shipmentData;
        foreach ($shipmentData['recipient'] as $key => $data) {
            if (empty($data)) {
                unset($newShipmentData['recipient'][$key]);
            }
        }
        $shipmentData = $newShipmentData;

        // Build data for shipment option id
        $selectedShippingOption = [
            'shipmentOptionId' =>
                (string) $shipmentOptionId,
        ];

        // Build data for selected add-ons
        if (!empty($selectedAddons)) {
            $shipmentDataSelectedAddons =
                [];
            foreach ($selectedAddons as $selectedAddonKey => $selectedAddonValue) {
                if (
                    !empty($selectedAddonValue['value'])
                    && is_string($selectedAddonValue['value'])
                ) {
                    $shipmentDataSelectedAddons[] = [
                        'data' => [
                            'text' =>
                                $selectedAddonValue['value'],
                        ],
                        'id' =>
                            (string) $selectedAddonKey,
                        'selected' =>
                            true,
                    ];
                } else {
                    $shipmentDataSelectedAddons[] = [
                        'id' =>
                            (string) $selectedAddonKey,
                        'selected' =>
                            true,
                    ];
                }
            }
            $selectedShippingOption['additionalServices'] =
                $shipmentDataSelectedAddons;
        }

        // Build data for selected service point id
        if (!empty($selectedServicePointId)) {
            $selectedShippingOption['servicePointId'] =
                $selectedServicePointId;
        }

        $shipmentData['selectedShippingOption'] =
            $selectedShippingOption;

        return $shipmentData;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function updateOrderShipmentReservation(
        int $orderId,
        string $externalShipmentId,
        string $shipmentOptionId,
        string $selectedServicePointId,
        array $selectedAddons
    ): bool {
        $shipmentData = $this->generateShipmentDataForOrder(
            $orderId,
            $shipmentOptionId,
            $selectedServicePointId,
            $selectedAddons
        );

        $shipment = $this->getOrderShipmentFromExternalShipmentId(
            $orderId,
            $externalShipmentId
        );
        if (!$shipment) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Failed to find internal shipment for order %1 external shipment id! %2',
                    $orderId,
                    $externalShipmentId
                )
            );
        }

        // Set weight on shipment to same weight as previous shipment
        $totalWeightInGrams = 0;
        foreach ($shipment->getPackages() as $package) {
            if (
                !empty($package['params'])
                && is_array($package['params'])
                && !empty($package['params']['weight'])
                && !empty($package['params']['weight_units'])
            ) {
                // Convert package-weight to grams
                $weightInGrams = 0;
                $packageWeight =
                    (float) $package['params']['weight'];
                if ($package['params']['weight_units'] === 'POUND') {
                    $weightInGrams =
                        $packageWeight * 453.5924;
                } elseif ($package['params']['weight_units'] === 'KILOGRAM') {
                    $weightInGrams =
                        $packageWeight * 1000;
                } else {
                    throw new \Magento\Framework\Exception\LocalizedException(
                        __(
                            'Unexpected weight unit %1!',
                            $package['params']['weight_units']
                        )
                    );
                }
                $totalWeightInGrams +=
                    (int) $weightInGrams;
            }
        }
        $shipmentData['weightInGrams'] =
            (int) $totalWeightInGrams;

        try {
            $this->metaHelper->updateReservedShipment(
                $orderId,
                $externalShipmentId,
                $shipmentData
            );
            return true;
        } catch (\Exception $e) {
            return false;
        }
    }

    public function deleteReservation(
        int $orderId,
        string $shipmentId
    ): void {
        $this->metaHelper->deleteReservedShipment(
            $orderId,
            $shipmentId
        );
        $this->deleteShipment(
            $orderId,
            $shipmentId
        );
    }
}
